/*
 * buffer.c
 *
 * routines of input, and formatting according to the styles
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97 Akim Demaille, Miguel Santana
 * $Id: buffer.c,v 1.34 1997/03/03 16:51:33 demaille Exp $
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "main.h"
#include "styles.h"
#include "buffer.h"
#include "style-utils.h"
#include "routines.h"
#include "getline.h"
#include "jobs.h"
#include "postscript.h"

/* The buffer structure, which encloses a lower case version to
 * speed up case insensitive parsing */
typedef struct {
  ustring content;
  ustring lowerCaseContent;
  size_t allocsize;
  int size;
  int curr;
} Buffer;

/* Description of the current style used */
STYLE theStyle = plain_style;
int translateSymbols = FALSE;

/* The alphabet used */
int inAlphabet1[256];
int inAlphabet2[256];

/* Sorted keywords */
keyword * sorted_keywords = NULL;
int num_keywords = 0;
int key_min[256], key_max[256];

/* Sorted regulars and specials */
symbol * sorted_regulars = NULL;
int num_regulars = 0;
int reg_min[256], reg_max[256];
symbol * sorted_specials = NULL;
int num_specials = 0;
int spe_min[256], spe_max[256];

/* Sequences ready to interpret */
matching_sequence * matching_sequences = NULL;
int num_sequences = 0;

/*
 * Use the information offered by main.c
 */
extern print_job * job;


/****************************************************************/
/*		Initialization of the pretty-printing	       	*/
/****************************************************************/
/*
 * Set up the variables depending on the current language
 */
void
setStyle (STYLE style)
{
  int c;

  if (style != plain_style && theStyle != style) {
    /* Set the alphabet tables */
    ustring list;
    
    for (c = 0 ; c <= 255 ; c++) {
      inAlphabet1[c] = FALSE;
      inAlphabet1[c] = FALSE;
      key_min[c] = reg_min[c] = spe_min[c] = -1;
      key_max[c] = reg_max[c] = spe_max[c] = -2;
    }
    
    for (list = (ustring) languages[style].first_characters ; 
	 *list ;  list += 2)
      for (c = *list ; c <= *(list+1) ; c++)
	inAlphabet1[c] = TRUE;
    
    for (list = (ustring) languages[style].following_characters ; 
	 *list ;  list += 2)
      for (c = *list ; c <= *(list+1) ; c++)
	inAlphabet2[c] = TRUE;
    
    /* Get a sorted version of the keywords */
    num_keywords = sort_keywords (&sorted_keywords, style, key_min, key_max);
    
    /* prepare the matching sequences */
    num_sequences = set_sequences (&matching_sequences, style);
  }
  theStyle = style;
}

void
setSymbolTranslation(int on)
{
  static STYLE last_style = plain_style;

  /* Make sure that symbol translation is always enabled with prescript,
   * and never in plain style */
  translateSymbols = ((theStyle != plain_style)
		      && (on
			  || (languages[theStyle].style == prescript_style)));

  /* If symbol translation is enabled and the style is not the last
   * style used, sort the arrays */
  if (translateSymbols && theStyle != last_style) {
    num_regulars = 
      sort_regulars (&sorted_regulars, theStyle, reg_min, reg_max);
    num_specials = 
      sort_specials (&sorted_specials, theStyle, spe_min, spe_max);
  }

  last_style = theStyle;
}

/****************************************************************/
/*		pretty printing service routines	       	*/
/****************************************************************/
/*
 * return TRUE if string begins with a sub-string in array
 */
inline static int 
is_in_pattern_array (ustring string, const uchar ***array)
{
  while (**array) {
    if (ustrprefix (**array, string))
      return TRUE;
    (*array) ++;
  }
  return FALSE;
}

/*
 * Does a matching_string matches the buffer?
 */
inline static int
match (matching_string mstr, Buffer * buffer)
{
  ustring string;
  
  string=(languages[theStyle].sensitiveness == CASE_SENSITIVE ?
	  buffer->content : buffer->lowerCaseContent) + buffer->curr;
  
  return (ustrprefix (mstr.string, string)
	  && (!mstr.at_start || buffer->curr == 0)
	  && (!mstr.at_end
	      || (buffer->curr + ustrlen (mstr.string) == buffer->size - 1)));
}

/*
 * Find where STRING is the start of a sequence
 */
inline static matching_sequence *
is_in_sequence_array (Buffer * buffer, matching_sequence * array, int curr)
{
  int i;

  for (i = 0 ; i < num_sequences ; i++)
    if (match (matching_sequences[i].open, buffer))
      return &matching_sequences[i];
  
  return NULL;
}

/*
 * puts in bufres the longuest word in buffer starting at curr
 */
static void 
read_word (Buffer * buffer, FONT* font, ustring bufres)
{
  int i=0;

  bufres[i++] = *(buffer->content + (buffer->curr++));
  while (inAlphabet2[*(buffer->content + (buffer->curr))])
    bufres[i++] = *(buffer->content + (buffer->curr++));;
  
  bufres[i] = NUL;
}


/****************************************************************/
/*			lexical analysis routines	       	*/
/****************************************************************/
/*
 * puts in bufres a keyword in buffer starting at curr, returns TRUE
 *	returns FALSE if not a keyword
 */
static int 
is_keyword (Buffer * buffer, FONT* font, ustring bufres)
{
  int i;
  uchar* string;
  
  string=(languages[theStyle].sensitiveness == CASE_SENSITIVE ?
	  buffer->content : buffer->lowerCaseContent) + buffer->curr;
  
  /* Return the biggest keyword that prefixes `string' such that
   * string[strlen(keyword)] is not in alphabet2 */
  for (i = key_max [*string] ; key_min [*string] <= i ; i--) {
    if (ustrprefix(sorted_keywords[i].theKeyword, string)
	&& !inAlphabet2[string[ustrlen(sorted_keywords[i].theKeyword)]])
      {/* Success */
	int length = ustrlen(sorted_keywords[i].theKeyword);
	ustrncpy(bufres, buffer->content+buffer->curr, length);
	bufres[length] = NUL;
	buffer->curr += length;
	*font = sorted_keywords[i].theFont;
	return TRUE;
      }
  }
  /* We found no keyword */
  return FALSE;
}

	    
/* return TRUE if string begins with a regular symbol
 * the converted symbol is in bufres
 */
static int
is_regular (Buffer * buffer, FONT* font, ustring bufres)
{
  int i;
  ustring string;
  
  string=(languages[theStyle].sensitiveness == CASE_SENSITIVE ?
	  buffer->content : buffer->lowerCaseContent) + buffer->curr;
  
  /* Return the biggest keyword that prefixes `string' such that
   * string[strlen(keyword)] is not in alphabet2 */
  for (i = reg_max [*string] ; reg_min [*string] <= i ; i--) {
    if (ustrprefix(sorted_regulars[i].theKeyword, string)
	&& !inAlphabet2[string[ustrlen(sorted_regulars[i].theKeyword)]])
      {
	ustrcpy(bufres, sorted_regulars[i].theSymbol);
	buffer->curr += ustrlen(sorted_regulars[i].theKeyword);
	*font = sorted_regulars[i].theFont;
	return TRUE;
      }
  }

  /* We found no regular matching */
  return FALSE;
}


/* return TRUE if string begins with a special symbol
 * list : types of comment for the current style
 */
static int 
is_special (Buffer * buffer, FONT* font, ustring bufres)
{
  int i;
  ustring string;
  
  string = buffer->content + buffer->curr;
  
  /* Then, return the biggest keyword that prefixes `string' */
  for (i = spe_max [*string] ; spe_min [*string] <= i ; i--) {
    if (ustrprefix(sorted_specials[i].theKeyword, string)) 
      {
	ustrcpy(bufres, sorted_specials[i].theSymbol);
	buffer->curr += ustrlen(sorted_specials[i].theKeyword);
	*font = sorted_specials[i].theFont;
	return TRUE;
      }
  }

  /* We found no special symbol matching string */
  return FALSE;
}


/*
 * return TRUE if string begins with an "verbatim" sequence
 */
static int 
is_verbatim (Buffer * buffer, ustring bufres)
{
  /* list : "verbatim" sequences, should be written straightly when
   * outside a sequence (e.g. in ada ''' must be written such as) */
  const uchar** list = languages[theStyle].verbatims;

  if (is_in_pattern_array(buffer->content+buffer->curr, &list)) {
    ustrcpy(bufres, *(list));
    buffer->curr += ustrlen(*list);
    return TRUE;
  }
  return FALSE;
}

/*
 * return TRUE if string begins with an "escape" sequence
 */
static int 
is_escape (Buffer * buffer, ustring bufres)
{
  /* list : "escape" sequences, should be written straightly when in a
     sequence (e.g. \" in a C string)  */
  const uchar** list = languages[theStyle].escapes;

  if (is_in_pattern_array(buffer->content+buffer->curr, &list)) {
    ustrcpy(bufres, *(list));
    buffer->curr += ustrlen(*list);
    return TRUE;
  }
  return FALSE;
}

/* 
 * return TRUE if buffer+*curr begins with a sequence
 */
static int
is_sequence_begin(Buffer * buffer, FONT* begin_font, 
                  matching_sequence ** currentSequence, ustring bufres)
{
  matching_sequence * list = matching_sequences;

  if ((*currentSequence = is_in_sequence_array(buffer, list, buffer->curr))) {
    int delta = ustrlen((*currentSequence)->open.string);
    /* printed opening string */
    ustrncpy(bufres, buffer->content+buffer->curr, delta);
    bufres[delta] = NUL;
    /* opening string in the source */
    buffer->curr += delta;
    /* font for the opening string */
    *begin_font = (*currentSequence)->open.font;
    
    return TRUE;
  }
  return FALSE;
}

/* 
 * return TRUE if buffer+*curr begins with with end sequence pattern
 */
static int 
is_sequence_end (Buffer * buffer, FONT* font, 
		 matching_sequence * currentSequence, ustring bufres)
{
  int i;
  
  for (i = 0 ; i < currentSequence->nb ; i++) {
    if (match (currentSequence->close[i], buffer)) {
      /* Don't foget to include the \n (or \r), when matching
       * which a $ */
      int delta = (ustrlen(currentSequence->close[i].string)
		   + (currentSequence->close[i].at_end ? 1 : 0));
      ustrncpy(bufres, buffer->content+buffer->curr, delta);
      bufres[delta] = NUL;
      buffer->curr += delta;
      *font = currentSequence->close[i].font;
      return TRUE;
    }
  }
  return FALSE;
}

/****************************************************************/
/*		 	Buffer Service routines	       		*/
/****************************************************************/
/* getBuffer
 *   uses GNU getline, and if the language is case insensitive,
 *   build a lower case version of the buffer
 */
static void
getBuffer(Buffer * buffer, char delim)
{
  buffer->size = 
    getdelim ((char **) &(buffer->content), &(buffer->allocsize), 
	      delim, stdin);
  
  /* If the eol char is preceded by a \f, then just forget the eol,
   * so that there won't be a blank line at the top of the next page */
  if ((buffer->size >= 2) 
      && buffer->content [buffer->size - 2] == '\f') {
    buffer->content [--(buffer->size)] = '\0';
  }

  if (theStyle != plain_style
      && languages[theStyle].sensitiveness == CASE_INSENSITIVE) {
    int i;
    buffer->lowerCaseContent = 
      xrealloc (buffer->lowerCaseContent, buffer->allocsize);
    for (i=0 ; i <= buffer->size ; i++)
      buffer->lowerCaseContent[i] = tolower(buffer->content[i]);
  }
  buffer->curr = 0;
}

#define isEmptyBuffer(Buf) (Buf.curr >= Buf.size)

/* - returns a char if no keyword are recocognized, or EOF, or
 *   IS_STRING if returns a string. Then the result is in the global
 * bufres */
static int 
mygetstring (FONT *font, ustring bufres)
{
  static Buffer buffer; /* = { UNULL, UNULL, 0, 0};*/
  /* are we currently in a sequence ? */
  static int in_sequence = FALSE; 
  static int return_to_courier = FALSE;
  /* font in a sequence, and closing symbol's font */
  static matching_sequence * currentSequence = NULL;

  int c;
  
  if (isEmptyBuffer(buffer)) {
    getBuffer(&buffer, encodings[job->encoding].new_line);
    /* On more line read */
    job->jobs->linenum++;

    if (buffer.size == -1) {
      /* end of file: reset values */
      /* If this is a new file, it must not depend on the trailling
       * parameters of the previous file */
      in_sequence = FALSE;
      return_to_courier = FALSE;
      return EOF;
    }
  }
  
  if (theStyle != plain_style) {
    if (return_to_courier) {
      return_to_courier = FALSE;
      *font = COURIER;
    }
    if (in_sequence) {
      *font = currentSequence->in_font;
      /* verbatim: not converted _anywhere_ */
      if (is_verbatim (&buffer, bufres))
	return IS_STRING;
      /* escape: not converted when in a sequence */
      if (is_escape (&buffer, bufres))
	return IS_STRING;
      /* end of sequence ? */
      if (is_sequence_end (&buffer, font, currentSequence, bufres))
	{
	  in_sequence = FALSE;
	  return_to_courier = TRUE;
	  return IS_STRING;
	}
    } else { /* (not in sequence) */
      /* whether we are in a string or not, verbatims are copied */
      if (is_verbatim (&buffer, bufres))
	return IS_STRING;
      if (is_sequence_begin (&buffer, font, &currentSequence, bufres))
	{
	  in_sequence = TRUE;
	  return IS_STRING;
	}
      else if (inAlphabet1[buffer.content[buffer.curr]]) {  
	/* we are in a word since this was a char belonging to the
	 * first alphabet */
	if (translateSymbols && is_regular (&buffer, font, bufres)) {
	  return_to_courier = TRUE;
	  return IS_STRING;
	} else if (is_keyword (&buffer, font, bufres))
	  return IS_STRING;
	else if (translateSymbols && is_special (&buffer, font, bufres)) {
	  return_to_courier = TRUE;
	  return IS_STRING;
	} else {
	  /* since some characters may be used inside an identifier
	   * (eg, x' = x in claire) but can also be used to open
	   * a special sequence (eg, 'x' in claire), then we must read
	   * the whole word, and print in.
	   */
	  read_word (&buffer, font, bufres);
	  return IS_STRING;
	}
      } else if (translateSymbols && is_special (&buffer, font, bufres)) {
	return_to_courier = TRUE;
	return IS_STRING;
      } else
	*font = COURIER;
    }
    return buffer.content[(buffer.curr)++];
  } else {
    /* This is plain style */
    *font = COURIER;
    c = buffer.content[(buffer.curr)++];
    
    /* Check if it is a bold sequence (nroff'ed pages style) */
    if (buffer.content[buffer.curr] == '\b') {
      if (c	== buffer.content[buffer.curr+1] &&
	  '\b'	== buffer.content[buffer.curr+2] &&
	  c	== buffer.content[buffer.curr+3] &&
	  '\b'	== buffer.content[buffer.curr+4] &&
	  c	== buffer.content[buffer.curr+5])
	{
	  *font = BOLD;
	  buffer.curr += 6;
	}
      /* Return the first character of the sequence */
    }
    return c;
  }
}

/*
 * read a new char of input. 
 */
void
print_postscript (print_job * job, STYLE lang)
{
  FONT font = COURIER;
  FONT new_font = COURIER;
  FONT true_font = COURIER;
  /* result of mygetc when not char */
  uchar bufres[BUFFER_SIZE+1];
  /* To grab the encoding switching instruction */
  uchar bufenc[BUFFER_SIZE+1] = "";
  int grabbing_encoding = FALSE;

  int c=0;
  int total_sheets;
  
  while ((c = mygetstring(&new_font, bufres)) != EOF) {
    /* Is a new font ? */
    if (font != new_font) {
      /* Reset dynamic markers */
      if (new_font & TAG1)
	*job->tag1 = NUL;
      if (new_font & TAG2)
	*job->tag2 = NUL;
      if (new_font & TAG3)
	*job->tag3 = NUL;
      if (new_font & TAG4)
	*job->tag4 = NUL;
      if (grabbing_encoding && (!(new_font & ENCODING_TAG))) {
	/* Grabbing of the encoding name is completed */
	ENCODING newenc = select_encoding(bufenc);
	if (newenc == NOENCODING)
	  error (_("unknown encoding: `%s', ignored"), bufenc);
	else
	  ps_switch_encoding (job, newenc);
	* bufenc = NUL;
	grabbing_encoding = FALSE;
      }
      font = new_font;
      true_font = font & FONT_MASK;
    }
    /* Grab markers if there are */
    switch (c) {
    case IS_STRING:
      if (font & TAG1)
	ustrcat(job->tag1, bufres);
      if (font & TAG2)
	ustrcat(job->tag2, bufres);
      if (font & TAG3)
	ustrcat(job->tag3, bufres);
      if (font & TAG4)
	ustrcat(job->tag4, bufres);
      break;
    default:
      if (font & TAG1)
	USTRCCAT(job->tag1, c);
      if (font & TAG2)
	USTRCCAT(job->tag2, c);
      if (font & TAG3)
	USTRCCAT(job->tag3, c);
      if (font & TAG4)
	USTRCCAT(job->tag4, c);
      /* Grab the dynamic encodings */
      if (font & ENCODING_TAG) {
	grabbing_encoding = TRUE;
	USTRCCAT(bufenc, c);
      }
      break;
    } 
  
    /* if the font is INVISIBLE, forget everything */
    /* if strip, skip (strong_)comments */
    if ((font & INVISIBLE)
	|| ((job->strip == 1 || job->strip == 3)
	    && font == COMMENT)
	|| ((job->strip == 2 || job->strip == 3)
	    && font == COMMENT_STRONG))
      continue;
    
    /* Return what's new */
    switch (c) {
    case IS_STRING:
      ps_print_string (job, bufres, true_font);
      break;
    default:
      ps_print_char (job, c, true_font);
      break;
    }
  }
  ps_end_file (job);
  
  /* Print the number of pages and sheets printed */
  total_sheets = job->jobs->sheets;
  if (job->rectoverso)
    total_sheets = (total_sheets + 1) / 2;
  if (job->jobs->pages == 1) 
    /* 1 page on 1 sheet */
    message (0, _("[%s(%s): 1 page on 1 sheet]\n"), 
	     job->jobs->name,
	     lang == plain_style ? _("plain") : languages[lang].name);
  else if (total_sheets == 1)
    /* several pages on 1 sheet */
    message (0, _("[%s(%s): %d pages on 1 sheet]\n"), job->jobs->name,
	     lang == plain_style ? _("plain") : languages[lang].name,
	     job->jobs->pages);
  else
    /* several sheets */
    message (0, _("[%s(%s): %d pages on %d sheets]\n"), job->jobs->name,
	     lang == plain_style ? _("plain") : languages[lang].name,
	     job->jobs->pages, total_sheets);
}
